"""
The imagecalc module contains functions for performing a number of calculating on images.
"""

# import the C++ extension into this level
from ._imagecalc import *

haveGDALPy = True
try:
    import osgeo.gdal as gdal
except ImportError:
    haveGDALPy = False

import rsgislib
import numpy

haveRIOS = True
try:
    from rios import applier
    from rios import cuiprogress
except ImportError as riosErr:
    haveRIOS = False

# define our own classes
class BandDefn(object):
    """
Create a list of these objects to pass to the bandMath function as the 'bands' parameter.
"""
    def __init__(self, bandName=None, fileName=None, bandIndex=None):
        self.bandName = bandName
        self.fileName = fileName
        self.bandIndex = bandIndex


class StatsSummary:
    """ 
This is passed to the imagePixelColumnSummary function 
"""
    def __init__(self, min=0.0, max=0.0, sum=0.0, median=0.0, stdDev=0.0, mean=0.0,
                 calcMin=False, calcMax=False, calcSum=False, calcMean=False, calcStdDev=False, calcMedian=False):
        self.min = min
        self.max = max
        self.sum = sum
        self.mean = mean
        self.stdDev = stdDev
        self.median = median
        self.calcMin = calcMin
        self.calcMax = calcMax
        self.calcSum = calcSum
        self.calcMean = calcMean
        self.calcStdDev = calcStdDev
        self.calcMedian = calcMedian



class ImageBandRescale(object):
    """
Data structure for rescaling information for rescaleImgPxlVals function.
* band - specified image band (band numbering starts at 1).
* inMin - the input image band minimum value for rescaling. 
* inMax - the input image band maximum value for rescaling.
* inNoData - no data value for the input image band.
* outMin - the output image band minimum value for rescaling. 
* outMax - the output image band maximum value for rescaling.
* outNoData - no data value for the output image band.
"""
    def __init__(self, band=0, inMin=0.0, inMax=0.0, inNoData=0, outMin=0.0, outMax=0.0, outNoData=0.0):
        """
        * band - specified image band (band numbering starts at 1).
        * inMin - the input image band minimum value for rescaling. 
        * inMax - the input image band maximum value for rescaling.
        * inNoData - no data value for the input image band.
        * outMin - the output image band minimum value for rescaling. 
        * outMax - the output image band maximum value for rescaling.
        * outNoData - no data value for the output image band.
        """
        self.band = band
        self.inMin = inMin
        self.inMax = inMax
        self.inNoData = inNoData
        self.outMin = outMin
        self.outMax = outMax
        self.outNoData = outNoData
        
    def __str__(self):
        strVal = 'Band ' + str(self.band) + " ["+str(self.inMin)+", "+str(self.inMax)+", "+str(self.inNoData)+"] ["+str(self.outMin)+", "+str(self.outMax)+", "+str(self.outNoData)+"]"
        return strVal
        
    def __repr__(self):
        strVal = 'Band ' + str(self.band) + " ["+str(self.inMin)+", "+str(self.inMax)+", "+str(self.inNoData)+"] ["+str(self.outMin)+", "+str(self.outMax)+", "+str(self.outNoData)+"]"
        return strVal


def calcDist2ImgVals(inputValsImg, outputDistImg, pxlVals, valsImgBand=1, gdalformat='KEA', maxDist=None, noDataVal=None, unitGEO=True):
    """ 
A function to calculate the distance to the nearest pixel value with one of the specified values.

Where:

* inputValsImg is a string specifying the input image file.
* outputDistImg is a string specfiying the output image file.
* pxlVals is a number of list of numbers specifying the features to which the distance from should be calculated.
* valsImgBand is an integer specifying the image band of the input image to be used (Default = 1).
* gdalformat is a string specifying the output image format (Default = KEA)
* maxDist is a number specifying the maximum distance to be calculated, if None not max value is used (Default = None).
* noDataVal is the no data value in the input image for which distance should not be calculated for (Default = None; None = no specified no data value).
* unitGEO is a boolean specifying the output distance units. True = Geographic units (e.g., metres), False is in Pixels (Default = True).

Example::

    import rsgislib.imagecalc
    cloudsImg = 'LS5TM_20110701_lat52lon421_r24p204_clouds.kea'
    dist2Clouds = 'LS5TM_20110701_lat52lon421_r24p204_distclouds.kea'
    # Pixel value 1 == Clouds
    # Pixel value 2 == Cloud Shadows
    rsgislib.imagecalc.calcDist2ImgVals(cloudsImg, dist2Clouds, pxlVals=[1,2])
    
"""
    # Check gdal is available
    if not haveGDALPy:
        raise ImportError("The GDAL python bindings are required for "
                          "calcDist2ImgVals function could not be imported")
    import rsgislib.imageutils
    
    haveListVals = False
    if type(pxlVals) is list:
        haveListVals = True
    
    proxOptions = []
    
    if maxDist is not None:
        proxOptions.append('MAXDIST='+str(maxDist))
    if noDataVal is not None:
        proxOptions.append('NODATA='+str(noDataVal))
    if unitGEO:
        proxOptions.append('DISTUNITS=GEO')
    else:
        proxOptions.append('DISTUNITS=PIXEL')
    
    if haveListVals:
        strVals = ''
        first = True
        for val in pxlVals:
            if first:
                strVals = str(val)
                first = False
            else:
                strVals = strVals + "," + str(val)
        proxOptions.append('VALUES='+strVals)    
    else:
        proxOptions.append('VALUES='+str(pxlVals))
        
    valsImgDS = gdal.Open(inputValsImg, gdal.GA_ReadOnly)
    valsImgBand = valsImgDS.GetRasterBand(valsImgBand)
    rsgislib.imageutils.createCopyImage(inputValsImg, outputDistImg, 1, 0.0, gdalformat, rsgislib.TYPE_32FLOAT)
    distImgDS = gdal.Open(outputDistImg, gdal.GA_Update)
    distImgBand = distImgDS.GetRasterBand(1)
    gdal.ComputeProximity(valsImgBand, distImgBand, proxOptions, callback=gdal.TermProgress)
    distImgBand = None
    distImgDS = None
    classImgBand = None
    classImgDS = None


def _computeProximityArrArgsFunc(argVals):
    """
This function is used internally within calcDist2Classes for the multiprocessing Pool
"""
    import rsgislib.imageutils
    classImgDS = gdal.Open(argVals[0], gdal.GA_ReadOnly)
    classImgBand = classImgDS.GetRasterBand(1)
    rsgislib.imageutils.createCopyImage(argVals[0], argVals[1], 1, argVals[3], argVals[4], rsgislib.TYPE_32FLOAT)
    distImgDS = gdal.Open(argVals[1], gdal.GA_Update)
    distImgBand = distImgDS.GetRasterBand(1)
    gdal.ComputeProximity(classImgBand, distImgBand, argVals[2], callback=gdal.TermProgress)
    distImgBand = None
    distImgDS = None
    classImgBand = None
    classImgDS = None



def calcDist2ImgValsTiled(inputValsImg, outputDistImg, pxlVals, valsImgBand=1, maxDist=1000, noDataVal=1000, gdalformat='KEA', unitGEO=True, tmpDIR='./tmp', tileSize=2000,  nCores=-1):
    """ 
A function to calculate the distance to the nearest pixel value with one of the specified values.

Where:

* inputValsImg is a string specifying the input image file.
* outputDistImg is a string specfiying the output image file.
* pxlVals is a number of list of numbers specifying the features to which the distance from should be calculated.
* valsImgBand is an integer specifying the image band of the input image to be used (Default = 1).
* gdalformat is a string specifying the output image format (Default = KEA)
* maxDist is a number specifying the maximum distance to be calculated, if None not max value is used (Default = None).
* noDataVal is the no data value in the input image for which distance should not be calculated for (Default = None; None = no specified no data value).
* unitGEO is a boolean specifying the output distance units. True = Geographic units (e.g., metres), False is in Pixels (Default = True).
* tmpDIR is a directory to be used for storing the image tiles and other temporary files - if not directory does not exist it will be created and deleted on completion (Default: ./tmp).
* tileSize is an int specifying in pixels the size of the image tiles used for processing (Default: 2000)
* nCores is the number of processing cores which are available to be used for this processing. If -1 all available cores will be used. (Default: -1)

Example::

    import rsgislib.imagecalc
    cloudsImg = 'LS5TM_20110701_lat52lon421_r24p204_clouds.kea'
    dist2Clouds = 'LS5TM_20110701_lat52lon421_r24p204_distclouds.kea'
    # Pixel value 1 == Clouds
    # Pixel value 2 == Cloud Shadows
    rsgislib.imagecalc.calcDist2ImgValsTiled(cloudsImg, dist2Clouds, pxlVals=[1,2])

"""
    
    # Check gdal is available
    if not haveGDALPy:
        raise ImportError("The GDAL python bindings are required for "
                          "calcDist2ImgValsTiled function could not be imported")
    import os.path
    import math
    import glob
    import shutil    
    from multiprocessing import Pool
    import rsgislib.imageutils
    
    haveListVals = False
    if type(pxlVals) is list:
        haveListVals = True
    
    tmpPresent = True
    if not os.path.exists(tmpDIR):
        print("WARNING: \'" + tmpDIR + "\' directory does not exist so creating it...")
        os.makedirs(tmpDIR)
        tmpPresent = False
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    if nCores <= 0:
        nCores = rsgisUtils.numProcessCores()
    
    uid = rsgisUtils.uidGenerator()
    
    xRes, yRes = rsgisUtils.getImageRes(inputValsImg)    
    if unitGEO:
        xMaxDistPxl = math.ceil(maxDist/xRes)
        yMaxDistPxl = math.ceil(maxDist/yRes)
    else:
        xMaxDistPxl = maxDist
        yMaxDistPxl = maxDist
                
    print("Max Dist Pxls X = {}, Y = {}".format(xMaxDistPxl, yMaxDistPxl))
    
    tileOverlap = xMaxDistPxl
    if yMaxDistPxl > xMaxDistPxl:
        tileOverlap = yMaxDistPxl
    
    imgTilesDIR = os.path.join(tmpDIR, 'ImgTiles_'+uid)
    imgTilesDIRPresent = True
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
        imgTilesDIRPresent = False
        
    imgTileBase = os.path.join(imgTilesDIR, 'ImgTile')
    rsgislib.imageutils.createTiles(inputValsImg, imgTileBase, tileSize, tileSize, tileOverlap, 0, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputValsImg), 'kea')
    imgTileFiles = glob.glob(imgTileBase+'*.kea')
    
    distTilesDIR = os.path.join(tmpDIR, 'DistTiles_'+uid)
    distTilesDIRPresent = True
    if not os.path.exists(distTilesDIR):
        os.makedirs(distTilesDIR)
        distTilesDIRPresent = False    
    
    proxOptions = []
    
    if maxDist is not None:
        proxOptions.append('MAXDIST='+str(maxDist))
    if noDataVal is not None:
        proxOptions.append('NODATA='+str(noDataVal))
    if unitGEO:
        proxOptions.append('DISTUNITS=GEO')
    else:
        proxOptions.append('DISTUNITS=PIXEL')
    
    if haveListVals:
        strVals = ''
        first = True
        for val in pxlVals:
            if first:
                strVals = str(val)
                first = False
            else:
                strVals = strVals + "," + str(val)
        proxOptions.append('VALUES='+strVals)    
    else:
        proxOptions.append('VALUES='+str(pxlVals))
    

    distTiles = []
    distTileArgs = []
    for tileFile in imgTileFiles:
        baseTileName = os.path.basename(tileFile)
        distTileFile = os.path.join(distTilesDIR, baseTileName)
        tileArgs = [tileFile, distTileFile, proxOptions, noDataVal, 'KEA']
        distTiles.append(distTileFile)
        distTileArgs.append(tileArgs)
    
    with Pool(nCores) as p:
        p.map(_computeProximityArrArgsFunc, distTileArgs)
            
    # Mosaic Tiles
    rsgislib.imageutils.createImageMosaic(distTiles, outputDistImg, 0, 0, 1, 1, gdalformat, rsgislib.TYPE_32FLOAT)
    rsgislib.imageutils.popImageStats(outputDistImg, usenodataval=True, nodataval=0, calcpyramids=True)
    
    for imgFile in distTiles:
        rsgisUtils.deleteFileWithBasename(imgFile)
    
    if not imgTilesDIRPresent:
        shutil.rmtree(imgTilesDIR, ignore_errors=True)
    else:
        for tileFile in imgTileFiles:
            rsgisUtils.deleteFileWithBasename(tileFile)
    
    if not distTilesDIRPresent:
        shutil.rmtree(distTilesDIR, ignore_errors=True)
    
    if not tmpPresent:
        shutil.rmtree(tmpDIR, ignore_errors=True)


def countPxlsOfVal(inputImg, vals=[0]):
    """
Function which counts the number of pixels of a set of values returning a list in the same order as the list of values provided.

Where:

* inputImg - the input image
* vals - is a list of pixel values to be counted

"""
    if len(vals) == 0:
        raise Exception('At least 1 value should be provided within the vals input varable.')
    numVals = len(vals)
    outVals = numpy.zeros(numVals,dtype=numpy.int64)
    
    from rios.imagereader import ImageReader

    reader = ImageReader(inputImg) 
    for (info, block) in reader:     
        for idx in range(numVals):
            outVals[idx] = outVals[idx] + (block == vals[idx]).sum()

    return outVals


def getPCAEigenVector(inputImg, pxlNSample, noData=None, outMatrixFile=None):
    """
A function which takes a sample from an input image and uses it to 
generate eigenvector for a PCA. Note. this can be used as input to rsgislib.imagecalc.pca

Where:

* inputImg - the image from which the random sample will be taken.
* pxlNSample - the sample to be taken (e.g., a value of 100 will sample every 100th pixel)
* noData - provide a no data value which is to be ignored during processing. If None then ignored (Default: None)
* outMatrixFile - path and name for the output rsgislib matrix file. If None file is not created (Default: None)

returns:

* array with the eigenvector
* array with the ratio of the explained variance

""" 
    from sklearn.decomposition import PCA
    import rsgislib.imageutils
    
    # Read input data from image file.
    X = rsgislib.imageutils.extractImgPxlSample(inputImg, pxlNSample, noData)
    
    print(str(X.shape[0]) + ' values were extracted from the input image.')
    
    pca = PCA()
    pca.fit(X)
    
    if outMatrixFile is not None:
        f = open(outMatrixFile, 'w')
        f.write('m='+str(pca.components_.shape[0])+'\n')
        f.write('n='+str(pca.components_.shape[1])+'\n')
        first = True
        for val in pca.components_.flatten():
            if first:
                f.write(str(val))
                first = False
            else:
                f.write(','+str(val))
        f.write('\n\n')
        f.flush()
        f.close()
    
    pcaComp = 1
    print("Prop. of variance explained:")
    for val in pca.explained_variance_ratio_:
        print('\t PCA Component ' + str(pcaComp) + ' = ' + str(round(val, 4)))
        pcaComp = pcaComp + 1

    return pca.components_, pca.explained_variance_ratio_


def performImagePCA(inputImg, outputImg, eigenVecFile, nComponents=None, pxlNSample=100, gdalformat='KEA', datatype=rsgislib.TYPE_32UINT, noData=None, calcStats=True):
    """
A function which performs a PCA on the input image.

Where:

* inputImg - the image from which the random sample will be taken.
* outputImg - the output image transformed using the calculated PCA
* eigenVecFile - path and name for the output rsgislib matrix file containing the eigenvector for the PCA.
* nComponents - the number of PCA compoents outputted. A value of None is all components (Default: None)
* pxlNSample - the sample to be taken (e.g., a value of 100 will sample every 100th pixel) (Default: 100)
* gdalformat - the output gdal supported file format (Default KEA)
* datatype - the output data type of the input image (Default: rsgislib.TYPE_32UINT)
* noData - provide a no data value which is to be ignored during processing. If None then ignored (Default: None)
* calcStats - Boolean specifying whether pyramids and statistics should be calculated for the output image. (Default: True)

"""
    import rsgislib.imageutils
    eigenVec, varExplain = rsgislib.imagecalc.getPCAEigenVector(inputImg, pxlNSample, noData, eigenVecFile)
    outNComp = varExplain.shape[0]
    if nComponents is not None:
        if nComponents > varExplain.shape[0]:
            raise Exception("You cannot output more components than the number of input image bands.")
        outNComp = nComponents
    
    rsgislib.imagecalc.pca(inputImg, eigenVecFile, outputImg, outNComp, gdalformat, datatype)
    if calcStats:
        usenodataval=False
        nodataval=0
        if noData is not None:
            usenodataval=True
            nodataval=noData
        rsgislib.imageutils.popImageStats(outputImg, usenodataval, nodataval, True)


def rescaleImgPxlVals(inputImg, outputImg, gdalformat, datatype, bandRescale, trim2Limits=True):
    """
Function which rescales an input image base on a list of rescaling parameters.

Where:

* inputImg - the input image
* outputImg - the output image file name and path (will be same dimensions as the input)
* gdalformat - the GDAL image file format of the output image file.
* bandRescale - list of ImageBandRescale objects
* trim2Limits - whether to trim the output to the output min/max values.

"""
    bandRescaleDict = dict()
    for rescaleObj in bandRescale:
        bandRescaleDict[rescaleObj.band-1] = rescaleObj
    
    numpyDT = numpy.float32
    if datatype == rsgislib.TYPE_8INT:
        numpyDT = numpy.int8
    elif datatype == rsgislib.TYPE_16INT:
        numpyDT = numpy.int16
    elif datatype == rsgislib.TYPE_32INT:
        numpyDT = numpy.int32
    elif datatype == rsgislib.TYPE_64INT:
        numpyDT = numpy.int64
    elif datatype == rsgislib.TYPE_8UINT:
        numpyDT = numpy.uint8
    elif datatype == rsgislib.TYPE_16UINT:
        numpyDT = numpy.uint16
    elif datatype == rsgislib.TYPE_32UINT:
        numpyDT = numpy.uint32
    elif datatype == rsgislib.TYPE_64UINT:
        numpyDT = numpy.uint64
    elif datatype == rsgislib.TYPE_32FLOAT:
        numpyDT = numpy.float32
    elif datatype == rsgislib.TYPE_64FLOAT:
        numpyDT = numpy.float64
    else:
        raise Exception('Datatype was not recognised.')
    
    infiles = applier.FilenameAssociations()
    infiles.image = inputImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    otherargs.rescaleDict = bandRescaleDict
    otherargs.trim = trim2Limits
    otherargs.numpyDT = numpyDT
    aControls = applier.ApplierControls()
    aControls.progress = cuiprogress.CUIProgressBar()
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    
    def _applyRescale(info, inputs, outputs, otherargs):
        """
        This is an internal rios function 
        """
        outputs.outimage = numpy.zeros_like(inputs.image, dtype=numpyDT)
        for idx in range(inputs.image.shape[0]):
            outputs.outimage[idx] = numpy.where(inputs.image[idx] == otherargs.rescaleDict[idx].inNoData, otherargs.rescaleDict[idx].outNoData, (((inputs.image[idx]-otherargs.rescaleDict[idx].inMin)/(inputs.image[idx]-otherargs.rescaleDict[idx].inMax - inputs.image[idx]-otherargs.rescaleDict[idx].inMin)) * (inputs.image[idx]-otherargs.rescaleDict[idx].outMax - inputs.image[idx]-otherargs.rescaleDict[idx].outMin)) + inputs.image[idx]-otherargs.rescaleDict[idx].outMin)
            if otherargs.trim:
                outputs.outimage[idx] = numpy.where((outputs.outimage[idx] != otherargs.rescaleDict[idx].outNoData) & (outputs.outimage[idx]<otherargs.rescaleDict[idx].outMin), otherargs.rescaleDict[idx].outMin, outputs.outimage[idx])
                outputs.outimage[idx] = numpy.where((outputs.outimage[idx] != otherargs.rescaleDict[idx].outNoData) & (outputs.outimage[idx]>otherargs.rescaleDict[idx].outMax), otherargs.rescaleDict[idx].outMax, outputs.outimage[idx])

    applier.apply(_applyRescale, infiles, outfiles, otherargs, controls=aControls)





#!/usr/bin/env python

"""
These utilities allow for a more 'intelligent tiling process to be carried out:

Example::

    import rsgislib
    from rsgislib.imageutils import tilingutils
    
    inputImage = 'LS5TM_20110428_sref_submask_osgb.kea'
    tileShp = 'LS5TM_20110428_sref_submask_osgb_tiles.shp'
    outTilesMaskBase = 'tilesmeta/LS5TM_20110428_sref_submask_osgb_tile'
    outTilesImgBase = 'tiles/LS5TM_20110428_sref_submask_osgb_tile'
    width = 2500
    height = 2500
    validDataThreshold = 0.3
    
    tilingutils.createMinDataTiles(inputImage, tileShp, width, height, validDataThreshold)
    tilingutils.createTileMaskImages(inputImage, tileShp, outTilesMaskBase)
    tilingutils.createTilesFromMasks(inputImage, outTilesMaskBase, outTilesImgBase, rsgislib.TYPE_16UINT, 'KEA')

"""
############################################################################
#  tilingutils.py
#
#  Copyright 2015 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 02/04/2015
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import sys
import math
import glob
import os.path
import shutil

import rsgislib
from rsgislib import segmentation
from rsgislib import rastergis
from rsgislib import imageutils
from rsgislib import vectorutils

haveGDALPy = True
try:
    import osgeo.gdal as gdal, ogr
except ImportError as gdalErr:
    haveGDALPy = False
    
haveRIOS = True
try:
    from rios import rat
except ImportError as riosErr:
    haveRIOS = False
    
haveNumpy = True
try:
    import numpy
except ImportError as numpyErr:
    haveNumpy = False

def createMinDataTiles(inputImage, outshp, outclumpsFile, width, height, validDataThreshold, maskIntersect=None, offset=False, force=True, tmpdir='tilestemp', inImgNoDataVal=0.0):
    """
A function to create a tiling for an input image where each tile has a minimum amount of valid data.

Where:

* inputImage is a string for the image to be tiled
* outshp is a string for the output shapefile the tiling will be written to (if None a shapefile won't be outputted).
* outclumpsFile is a string for the output image file containing the tiling
* width is an int for the width of the tiles
* height is an int for the height of the tiles
* validDataThreshold is a float (0-1) with the proportion of valid data needed within a tile.
* force is a boolean (default True) to delete the output shapefile if it already exists.
* tmpdir is a string with a temporary directory for temp outputs to be stored (they will be deleted) 
    * if tmpdir doesn't exist it will be created and then deleted during the processing.
* inImgNoDataVal is a float for providing the input image no data value (Default: 0.0)

"""
    tmpPresent = True
    if not os.path.exists(tmpdir):
        print("WARNING: tmpdir directory does not exist so creating it...")
        os.makedirs(tmpdir)
        tmpPresent = False
        
    inImgBaseName = os.path.splitext(os.path.basename(inputImage))[0]
    
    tileClumpsImage = os.path.join(tmpdir, inImgBaseName+'_tilesimg.kea')
        
    segmentation.generateRegularGrid(inputImage, tileClumpsImage, 'KEA', width, height, offset)
    rastergis.populateStats(tileClumpsImage, True, True)

    if not maskIntersect == None:
        bs = []
        bs.append(rastergis.BandAttStats(band=1, maxField='Mask'))
        rastergis.populateRATWithStats(maskIntersect, tileClumpsImage, bs)
        ratDS = gdal.Open(tileClumpsImage, gdal.GA_Update)
        MaskField = rat.readColumn(ratDS, "Mask")
        Selected = numpy.zeros_like(MaskField, dtype=int)
        Selected[MaskField == inImgNoDataVal] = 1
        rat.writeColumn(ratDS, "Selected", Selected)
        ratDS = None
        
        tileClumpsImageDropClumps = os.path.join(tmpdir, inImgBaseName+'_tilesimgdropped.kea')

        segmentation.dropSelectedClumps(tileClumpsImage, tileClumpsImageDropClumps, 'KEA', 'Selected')
        os.remove(tileClumpsImage)
        tileClumpsImage = tileClumpsImageDropClumps
    
         
    rastergis.populateRATWithPropValidPxls(inputImage, tileClumpsImage, "ValidPxls", inImgNoDataVal)
    
    ratDS = gdal.Open(tileClumpsImage, gdal.GA_Update)
    ValidPxls = rat.readColumn(ratDS, "ValidPxls")
    Selected = numpy.zeros_like(ValidPxls, dtype=int)
    NoDataClumps = numpy.zeros_like(ValidPxls, dtype=int)
    Selected[ValidPxls < validDataThreshold] = 1
    NoDataClumps[ValidPxls == 0] = 1
    Selected[ValidPxls == 0] = 0
    rat.writeColumn(ratDS, "Selected", Selected)
    rat.writeColumn(ratDS, "NoDataClumps", NoDataClumps)
    ratDS = None

    segmentation.mergeSegments2Neighbours(tileClumpsImage, inputImage, outclumpsFile, 'KEA', "Selected", "NoDataClumps")
    
    if not outshp is None:
        tilesDS = gdal.Open(outclumpsFile, gdal.GA_ReadOnly)
        tilesDSBand = tilesDS.GetRasterBand(1)
        
        dst_layername = os.path.splitext(os.path.basename(outshp))[0]
        #print(dst_layername)
        drv = ogr.GetDriverByName("ESRI Shapefile")
        
        if force and os.path.exists(outshp):
            drv.DeleteDataSource(outshp)
        
        dst_ds = drv.CreateDataSource( outshp )
        dst_layer = dst_ds.CreateLayer(dst_layername, srs = None )
        
        gdal.Polygonize( tilesDSBand, tilesDSBand, dst_layer, -1, [], callback=None )
        
        tilesDS = None
        dst_ds = None
    
    if not tmpPresent:
        shutil.rmtree(tmpdir, ignore_errors=True)
    else:
        os.remove(tileClumpsImage)

    
def createTileMaskImagesFromShp(inputImage, tileShp, tilesNameBase, tilesMaskDIR, tmpdir='tilestemp', imgFormat='KEA'):
    """
A function to create individual image masks from the tiles shapefile which can be
individually used to mask (using rsgislib mask function) each tile from the inputimage.

Where:

* inputImage is the input image being tiled.
* tileShp is a shapefile containing the shapefile tiles.
* tilesNameBase is the base file name for the tile masks
* tilesMaskDIR is the directory where the output images will be outputted
* tmpdir is a string with a temporary directory for temp outputs to be stored (they will be deleted)
    * If tmpdir doesn't exist it will be created and then deleted during the processing. 
"""
    
    tmpPresent = True
    if not os.path.exists(tmpdir):
        print("WARNING: tmpdir directory does not exist so creating it...")
        os.makedirs(tmpdir)
        tmpPresent = False
    
    shpTilesBase = os.path.join(tmpdir, tilesNameBase+'_tileshp')
    
    vectorutils.splitFeatures(tileShp, shpTilesBase, True)

    drv = ogr.GetDriverByName("ESRI Shapefile")
    shpFiles = glob.glob(shpTilesBase+"*.shp")
    idx = 1
    for shpFile in shpFiles:
        imgTileFile = os.path.join(tilesMaskDIR, tilesNameBase + str(idx) + '.kea')
        #print(imgTileFile)
        vectorutils.rasterise2Image(shpFile, inputImage, imgTileFile, imgFormat, shpAtt=None, shpExt=True)
        drv.DeleteDataSource(shpFile)
        idx = idx + 1
    
    if not tmpPresent:
        shutil.rmtree(tmpdir, ignore_errors=True)

def createTileMaskImagesFromClumps(clumpsImage, tilesNameBase, tilesMaskDIR, imgFormat='KEA'):
    """
A function to create individual image masks from the tiles shapefile which can be
individually used to mask (using rsgislib mask function) each tile from the inputimage.

Where:

* clumpsImage is an image file with RAT where each clump represented a tile region.
* tilesNameBase is the base file name for the tile masks
* tilesMaskDIR is the directory where the output images will be outputted
* imgFormat is the output image file format of the tile masks

"""
    outBaseImg = os.path.join(tilesMaskDIR, tilesNameBase)
    rsgisUtils = rsgislib.RSGISPyUtils()
    outImgExt = rsgisUtils.getFileExtension(imgFormat)[1:]
    rastergis.exportClumps2Images(clumpsImage, outBaseImg, True, outImgExt, imgFormat, 1)


def createTilesFromMasks(inputImage, tilesBase, tilesMetaDIR, tilesImgDIR, datatype, gdalformat):
    """
A function to apply the image tile masks defined in createTileMaskImages to the input image to extract the individual tiles.

Where:

* inputImage is the input image being tiled.
* tileMasksBase is the base path for the tile masks. glob will be used to find them with \*.kea added to the end.
* outTilesBase is the base file name for the tiles.

"""
    maskFiles = glob.glob(os.path.join(tilesMetaDIR, tilesBase+"*.kea"))

    idx = 1
    for maskFile in maskFiles:
        tileImage = os.path.join(tilesImgDIR, os.path.basename(maskFile))
        #print(tileImage)
        imageutils.maskImage(inputImage, maskFile, tileImage, gdalformat, datatype, 0, 0)
        imageutils.popImageStats(tileImage,True,0.,True)



"""
The segmentation module contains the segmentation functionality for RSGISLib.

A number of steps are required for the segmentation, for most users it is recommended to use the `runShepherdSegmentation`
helper function which will run all the required steps to generate a segmentation:

Example::

    from rsgislib.segmentation import segutils
    
    segutils.runShepherdSegmentation(inImage,
                                     outputClumps,
                                     tmpath='./',
                                     numClusters=60,
                                     minPxls=100,
                                     distThres=100,
                                     sampling=100, kmMaxIter=200)

Where 'inImage' is the input image (optionally masked and stretched) and 'outputClumps' is the output clumps file.

More information about the segmentation method is available in the following paper:

Daniel Clewley, Peter Bunting, James Shepherd, Sam Gillingham, Neil Flood, John Dymond, Richard Lucas, John Armston and Mahta Moghaddam. 2014. A Python-Based Open Source System for Geographic Object-Based Image Analysis (GEOBIA) Utilizing Raster Attribute Tables. Remote Sensing. Volume 6, Pages 6111-6135. http://www.mdpi.com/2072-4292/6/7/6111

"""

# import the C++ extension into this level
from ._segmentation import *
#import warnings
#
#
#def UnionOfClumps(outputimage, gdalformat, inputimagepaths, nodata):
#    """
#    Deprecated: is now 'unionOfClumps' (note starts with lower case 'u')
#    """
#    warnings.warn("'UnionOfClumps' has been renamed to 'unionOfClumps' (starts with lower case 'u'), please update your code.", Warning, stacklevel=2)
#    unionOfClumps(outputimage, gdalformat, inputimagepaths, nodata, False)
#
#
#def RMSmallClumpsStepwise(inputimage, clumpsimage, outputimage, gdalformat, stretchstatsavail, stretchstatsfile, storemean, processinmemory, minclumpsize, specThreshold):
#    """
#    Deprecated: is now 'rmSmallClumpsStepwise' (note starts with lower case 'rm')
#    """
#    warnings.warn("'RMSmallClumpsStepwise' has been renamed to 'rmSmallClumpsStepwise' (starts with lower case 'rm'), please update your code.", Warning, stacklevel=2)
#    rmSmallClumpsStepwise(inputimage, clumpsimage, outputimage, gdalformat, stretchstatsavail, stretchstatsfile, storemean, processinmemory, minclumpsize, specThreshold)
#    



#! /usr/bin/env python

############################################################################
#  tiledfilter.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The Remote Sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide functionality to perform image filtering using tiling
#           and multiple processing cores.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 14/05/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import rsgislib
from rsgislib import imageutils
from rsgislib import imagefilter

import glob
import os.path
import os
import shutil
import math
from multiprocessing import Pool
from abc import ABCMeta, abstractmethod

class RSGISAbstractFilter:
    """
    Abstract class for filter defining the interface to be used within the performTiledImgFilter function.
    Must define:
    
    * self.filterSize - size of the image filter (must be an odd number)
    * self.gdalformat - the output image file format
    * self.dataType - the output image data type (e.g., rsgislib.TYPE_16UINT)
    """
    __metaclass__ = ABCMeta
    
    def __init__(self):
        self.filterSize = 3
        self.gdalformat = 'KEA'
        self.dataType = rsgislib.TYPE_32FLOAT
    
    @abstractmethod
    def applyFilter(self, inputimage, outputImage): 
        """
        Abstract function through which the input image
        is filtered to produce the output image.
        
        * inputimage - is the name and path of the input image.
        * outputimage - is the name and path of the output image.
        """
        pass
    

    def getFilterSize(self):
        """
        return the size of the image filter
        """
        return self.filterSize
    
    def getFilterHSize(self):
        """
        return the hald size of the image filter (i.e., (filterSize-1)/2)
        """
        hSize = (self.filterSize-1)/2
        return hSize


class RSGISMedianFilter(RSGISAbstractFilter):
    """ A class to apply a median filter """
    
    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a median filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Median', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISMeanFilter(RSGISAbstractFilter):
    """ A class to apply a mean filter """
    
    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a mean filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Mean', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISMinFilter(RSGISAbstractFilter):
    """ A class to apply a min filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a min filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Min', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)
    

class RSGISMaxFilter(RSGISAbstractFilter):
    """ A class to apply a max filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a max filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Max', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISModeFilter(RSGISAbstractFilter):
    """ A class to apply a mode filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a mode filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Mode', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)

    
class RSGISStdDevFilter(RSGISAbstractFilter):
    """ A class to apply a standard deviation filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a std dev filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'StdDev', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)    


class RSGISRangeFilter(RSGISAbstractFilter):
    """ A class to apply a range filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a range filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Range', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)       
    

class RSGISMeanDiffFilter(RSGISAbstractFilter):
    """ A class to apply a mean difference filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a mean difference filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'MeanDiff', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISMeanDiffAbsFilter(RSGISAbstractFilter):
    """ A class to apply a mean absolute difference filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a mean absolute difference filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'MeanDiffAbs', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISTotalDiffFilter(RSGISAbstractFilter):
    """ A class to apply a total difference filter """
    
    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a total (i.e., sum) difference filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'TotalDiff', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISTotalDiffAbsFilter(RSGISAbstractFilter):
    """ A class to apply a total absolution difference filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a total absolute difference filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'TotalDiffAbs', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISCoeffOfVarFilter(RSGISAbstractFilter):
    """ A class to apply a coefficient of variance filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a coefficient of variance filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'CoeffOfVar', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISTotalFilter(RSGISAbstractFilter):
    """ A class to apply a total filter """
    
    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a total filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Total', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISNormVarFilter(RSGISAbstractFilter):
    """ A class to apply a normalised variance filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a normalised variance filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'NormVar', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISNormVarSqrtFilter(RSGISAbstractFilter):
    """ A class to apply a normalised variance square root filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a normalised variance square root filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'NormVarSqrt', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISNormVarLnFilter(RSGISAbstractFilter):
    """ A class to apply a normalised variance log filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a normalised variance natural log filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'NormVarLn', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISTextureVarFilter(RSGISAbstractFilter):
    """ A class to apply a texture variance filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a texture variance filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'TextureVar', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISKuwaharaFilter(RSGISAbstractFilter):
    """ A class to apply a kuwahara filter """

    def __init__(self, filterSize, gdalformat, dataType):
        self.filterSize = filterSize
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a kuwahara filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Kuwahara', fileEnding = '', size=self.filterSize) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISSobelFilter(RSGISAbstractFilter):
    """ A class to apply a sobel filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a sobel filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = '', option = 'xy') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)    


class RSGISSobelXFilter(RSGISAbstractFilter):
    """ A class to apply a sobel X filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a sobel filter in X axis to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = '', option = 'x') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISSobelYFilter(RSGISAbstractFilter):
    """ A class to apply a sobel Y filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a sobel filter in Y axis to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Sobel', fileEnding = '', option = 'y') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISPrewittFilter(RSGISAbstractFilter):
    """ A class to apply a Prewitt filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a Prewitt filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = '', option = 'xy') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)
    
  
class RSGISPrewittXFilter(RSGISAbstractFilter):
    """ A class to apply a Prewitt X filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType
   
    def applyFilter(self, inputimage, outputImage):
        """ Apply a Prewitt filter in X axis to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = '', option = 'x') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISPrewittYFilter(RSGISAbstractFilter):
    """ A class to apply a Prewitt Y filter """

    def __init__(self, gdalformat, dataType):
        self.filterSize = 3
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a Prewitt filter in Y axis to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Prewitt', fileEnding = '', option = 'y') )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISGaussianSmoothFilter(RSGISAbstractFilter):
    """ A class to apply a Gaussian smoothing filter """

    def __init__(self, filterSize, stddevX, stddevY, filterAngle, gdalformat, dataType):
        self.filterSize = filterSize
        self.stddevX = stddevX
        self.stddevY = stddevY
        self.filterAngle = filterAngle
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a Gaussian smoothing filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'GaussianSmooth', fileEnding = '', size=self.filterSize, stddevX=self.stddevX, stddevY=self.stddevY, angle=self.filterAngle) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)
    

class RSGISGaussian1stDerivFilter(RSGISAbstractFilter):
    """ A class to apply a Gaussian first derivative filter """

    def __init__(self, filterSize, stddevX, stddevY, filterAngle, gdalformat, dataType):
        self.filterSize = filterSize
        self.stddevX = stddevX
        self.stddevY = stddevY
        self.filterAngle = filterAngle
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a Gaussian first derivative filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Gaussian1st', fileEnding = '', size=self.filterSize, stddevX=self.stddevX, stddevY=self.stddevY, angle=self.filterAngle) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


class RSGISGaussian2ndDerivFilter(RSGISAbstractFilter):
    """ A class to apply a Gaussian second derivative filter """

    def __init__(self, filterSize, stddevX, stddevY, filterAngle, gdalformat, dataType):
        self.filterSize = filterSize
        self.stddevX = stddevX
        self.stddevY = stddevY
        self.filterAngle = filterAngle
        self.gdalformat = gdalformat
        self.dataType = dataType
    
    def applyFilter(self, inputimage, outputImage):
        """ Apply a Gaussian second derivative filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Gaussian2nd', fileEnding = '', size=self.filterSize, stddevX=self.stddevX, stddevY=self.stddevY, angle=self.filterAngle) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)

class RSGISLaplacianFilter(RSGISAbstractFilter):
    """ A class to apply a Laplacian filter """

    def __init__(self, filterSize, stddev, gdalformat, dataType):
        self.filterSize = filterSize
        self.stddev = stddev
        self.gdalformat = gdalformat
        self.dataType = dataType

    def applyFilter(self, inputimage, outputImage):
        """ Apply a Laplacian filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Laplacian', fileEnding = '', size=self.filterSize, stddev=self.stddev) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)
    
    
class RSGISLeeFilter(RSGISAbstractFilter):
    """ A class to apply a SAR Lee filter """

    def __init__(self, filterSize, nLooks, gdalformat, dataType):
        self.filterSize = filterSize
        self.nLooks = nLooks
        self.gdalformat = gdalformat
        self.dataType = dataType
 
    def applyFilter(self, inputimage, outputImage):
        """ Apply a Lee SAR filter to the specified input image."""
        outputImageBase, outExt = os.path.splitext(outputImage)
        outExt = outExt.replace(".", "").strip()
        filters = []
        filters.append(imagefilter.FilterParameters(filterType = 'Lee', fileEnding = '', size=self.filterSize, nLooks=self.nLooks) )
        imagefilter.applyfilters(inputimage, outputImageBase, filters, self.gdalformat, outExt, self.dataType)


def _performFilteringFunc(filterParams):
    """
    Clump an image with values provides as an array for use within a multiprocessing Pool
    """
    filterParams[2].applyFilter(filterParams[0], filterParams[1])


def performTiledImgFilter(inputImg, outputImg, filterInst, dataType=None, imgFormat='KEA', tmpDIR='tmp', width=2000, height=2000, nCores=-1):
    """
This function will perform a filtering of an input image where the input image will be tiled and the tiles executed on multiple processing cores. This function is primarily of use for larger images or when using very large filter windows otherwise the over head of tiling and mosaicking are not worth it.

* inputImg - is the file name and path for the input image file.
* outputImg - is the file name and path for the output image file.
* filterInst - is an instance of a filter class available within rsgislib.imagefilter.tiledfilter. 
* datatype - is the output image data type (e.g., rsgislib.TYPE_32FLOAT; Default is None). If None then data type of input image is used.
* imgformat - string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.
* tmpDIR - the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
* width - int for width of the image tiles used for processing (Default = 2000).
* height - int for height of the image tiles used for processing (Default = 2000).
* nCores - is an int specifying the number of cores to be used for clumping processing.

Example::

    import rsgislib
    from rsgislib.imagefilter import tiledfilter
    from rsgislib import imageutils
    
    inputImage = 'LandsatImg.kea'
    outputImage = 'LandsatImgMedianFilter.kea'
    
    medianFilter = tiledfilter.RSGISMedianFilter(7, "KEA", rsgislib.TYPE_16UINT)
    tiledfilter.performTiledImgFilter(inputImage, outputImage, medianFilter, width=2000, height=2000)
    imageutils.popImageStats(outputImage, usenodataval=False, nodataval=0, calcpyramids=True)
    
    """
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    if nCores <= 0:
        nCores = rsgisUtils.numProcessCores()
        
    uidStr = rsgisUtils.uidGenerator()
    if dataType == None:
        dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
    
    baseName = os.path.splitext(os.path.basename(inputImg))[0]+"_"+uidStr
    
    imgTilesDIR = os.path.join(tmpDIR, "imgTiles_"+uidStr)
    tilesFilterDIR = os.path.join(tmpDIR, "imgFilterTiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)

    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
        
    if not os.path.exists(tilesFilterDIR):
        os.makedirs(tilesFilterDIR)
    
    tileOverlap = math.ceil(filterInst.getFilterHSize())
    
    imageutils.createTiles(inputImg, tilesImgBase, int(width), int(height), int(tileOverlap), False, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), 'kea')
    imageTiles = glob.glob(tilesImgBase+"*.kea")    
    
    filterImgsVals = []
    for tile in imageTiles:
        tileBaseName = os.path.splitext(os.path.basename(tile))[0]
        filterTile = os.path.join(tilesFilterDIR, tileBaseName+'_filter.kea')
        filterImgsVals.append([tile, filterTile, filterInst])    
    
    with Pool(nCores) as p:
        p.map(_performFilteringFunc, filterImgsVals)
    
    imgFilterTiles = glob.glob(os.path.join(tilesFilterDIR,"*_filter.kea"))
    
    numOutBands = rsgisUtils.getImageBandCount(inputImg)
    
    imageutils.createCopyImage(inputImg, outputImg, numOutBands, 0, imgFormat, dataType)
    
    imageutils.includeImagesWithOverlap(outputImg, imgFilterTiles, int(tileOverlap))
    
    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesFilterDIR)
    if createdTmp:
        shutil.rmtree(tmpDIR)
        
def performTiledImgMultiFilter(inputImg, outputImgs, filterInsts, dataType=None, imgFormat='KEA', tmpDIR='tmp', width=2000, height=2000, nCores=-1):
    """
This function will perform the filtering using multiple filters of an input image where the input image will be tiled and the tiles executed on multiple processing cores. This function is primarily of use for larger images or when using very large filter windows otherwise the over head of tiling and mosaicking are not worth it.

* inputImg - is the file name and path for the input image file.
* outputImgs - is a list of file names and paths for the output image files - Note, must be the same length as filterInsts.
* filterInsts - is a list of filter instances of the classes available within rsgislib.imagefilter.tiledfilter  - Note, must be the same length as filterInsts.
* datatype - is the output image data type (e.g., rsgislib.TYPE_32FLOAT; Default is None). If None then data type of input image is used.
* imgformat - string with the GDAL image format for the output image (Default = KEA). NOTE. KEA is used as intermediate format internally and therefore needs to be available.
* tmpDIR - the temporary directory where intermediate files will be written (default is 'tmp'). Directory will be created and deleted if does not exist.
* width - int for width of the image tiles used for processing (Default = 2000).
* height - int for height of the image tiles used for processing (Default = 2000).
* nCores - is an int specifying the number of cores to be used for clumping processing.

Example::

    import rsgislib
    from rsgislib.imagefilter import tiledfilter
    from rsgislib import imageutils
    
    inputImage = 'LandsatImg.kea'
    outputImages = ['LandsatImgMedianFilter.kea', 'LandsatImgNormVarFilter.kea']
    
    filters = [tiledfilter.RSGISMedianFilter(7, "KEA", rsgislib.TYPE_16UINT), tiledfilter.RSGISNormVarFilter(7, "KEA", rsgisUtils.getRSGISLibDataTypeFromImg(inputImage))]
    tiledfilter.performTiledImgMultiFilter(inputImage, outputImages, filters, width=2000, height=2000)
    imageutils.popImageStats(outputImage, usenodataval=False, nodataval=0, calcpyramids=True)
    
    """
    
    if (len(outputImgs) != len(filterInsts)):
        raise rsgislib.RSGISPyException('The same number of filters and output images need to be provided.')
    
    numFilters = len(outputImgs)
    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    createdTmp = False
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        createdTmp = True
    
    if nCores <= 0:
        nCores = rsgisUtils.numProcessCores()
        
    uidStr = rsgisUtils.uidGenerator()
    if dataType == None:
        dataType = rsgisUtils.getRSGISLibDataTypeFromImg(inputImg)
    
    baseName = os.path.splitext(os.path.basename(inputImg))[0]+"_"+uidStr
    
    imgTilesDIR = os.path.join(tmpDIR, "imgTiles_"+uidStr)
    tilesFilterDIR = os.path.join(tmpDIR, "imgFilterTiles_"+uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)

    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    
    first = True
    tileOverlap = 0
    
    for filterInst in filterInsts:
        tmpOverlap = math.ceil(filterInst.getFilterHSize())
        if first:
            tileOverlap = tmpOverlap
            first = False
        elif tmpOverlap > tileOverlap:
            tileOverlap = tmpOverlap
        
    imageutils.createTiles(inputImg, tilesImgBase, int(width), int(height), int(tileOverlap), False, 'KEA', rsgisUtils.getRSGISLibDataTypeFromImg(inputImg), 'kea')
    imageTiles = glob.glob(tilesImgBase+"*.kea")    
    
    for i in range(numFilters):
        filterInst = filterInsts[i]
        outputImg = outputImgs[i]
        
        if not os.path.exists(tilesFilterDIR):
            os.makedirs(tilesFilterDIR)
        
        filterImgsVals = []
        for tile in imageTiles:
            tileBaseName = os.path.splitext(os.path.basename(tile))[0]
            filterTile = os.path.join(tilesFilterDIR, tileBaseName+'_filter.kea')
            filterImgsVals.append([tile, filterTile, filterInst])    
        
        with Pool(nCores) as p:
            p.map(_performFilteringFunc, filterImgsVals)
        
        imgFilterTiles = glob.glob(os.path.join(tilesFilterDIR,"*_filter.kea"))
        
        numOutBands = rsgisUtils.getImageBandCount(inputImg)
        
        imageutils.createCopyImage(inputImg, outputImg, numOutBands, 0, imgFormat, dataType)
        
        imageutils.includeImagesWithOverlap(outputImg, imgFilterTiles, int(filterInst.getFilterHSize()))
        
        shutil.rmtree(tilesFilterDIR)
        
    shutil.rmtree(imgTilesDIR)
    if createdTmp:
        shutil.rmtree(tmpDIR)




print("import: 'rsgislib'")
import rsgislib

print("import: 'rsgislib.imagecalibration'")
import rsgislib.imagecalibration

print("import: 'rsgislib.classification'")
import rsgislib.classification

print("import: 'rsgislib.elevation'")
import rsgislib.elevation

print("import: 'rsgislib.imagecalc'")
import rsgislib.imagecalc

print("import: 'rsgislib.imagefilter'")
import rsgislib.imagefilter

print("import: 'rsgislib.imageregistration'")
import rsgislib.imageregistration

print("import: 'rsgislib.imageutils'")
import rsgislib.imageutils

print("import: 'rsgislib.rastergis'")
import rsgislib.rastergis

print("import: 'rsgislib.segmentation'")
import rsgislib.segmentation

print("import: 'rsgislib.vectorutils'")
import rsgislib.vectorutils

print("import: 'rsgislib.zonalstats'")
import rsgislib.zonalstats

print("import: 'rsgislib.imagemorphology'")
import rsgislib.imagemorphology




"""
The imageutils module contains general utilities for applying to images.
"""
# Maintain python 2 backwards compatibility
from __future__ import print_function
# import the C++ extension into this level
from ._imageutils import *
import rsgislib 

import os.path
import math

import numpy

import osgeo.gdal as gdal
import osgeo.ogr as ogr
import osgeo.osr as osr

haveRIOS = True
try:
    from rios import applier
    from rios import cuiprogress
except ImportError as riosErr:
    haveRIOS = False

# define our own classes
class ImageBandInfo(object):
    """
Create a list of these objects to pass to the extractZoneImageBandValues2HDF function

* fileName - is the input image file name and path.
* name - is a name associated with this layer - doesn't really matter what you use but needs to be unique; this is used as a dict key in some functions. 
* bands - is a list of image bands within the fileName to be used for processing (band numbers start at 1).

"""
    def __init__(self, fileName=None, name=None, bands=None):
        """
        * fileName - is the input image file name and path.
        * name - is a name associated with this layer - doesn't really matter what you use but needs to be unique; this is used as a dict key in some functions. 
        * bands - is a list of image bands within the fileName to be used for processing (band numbers start at 1).
        """
        self.fileName = fileName
        self.name = name
        self.bands = bands
        
# define our own classes
class SharpBandInfo(object):
    """
Create a list of these objects to pass to the sharpenLowResBands function.

* band - is the band number (band numbering starts at 1).
* status - needs to be either rsgislib.SHARP_RES_IGNORE, rsgislib.SHARP_RES_LOW or rsgislib.SHARP_RES_HIGH
           lowres bands will be sharpened using the highres bands and ignored bands 
           will just be copied into the output image.
* name - is a name associated with this image band - doesn't really matter what you put in here. 

"""
    def __init__(self, band=None, status=None, name=None):
        """
        * band - is the band number (band numbering starts at 1).
        * status - needs to be either 'ignore', 'lowres' or 'highres' - lowres bands will be sharpened using the highres bands and ignored bands will just be copied into the output image.
        * name - is a name associated with this image band - doesn't really matter what you put in here. 
        """
        self.band = band
        self.status = status
        self.name = name

# Define Class for time series fill
class RSGISTimeseriesFillInfo(object):
    """
Create a list of these objects to pass to the fillTimeSeriesGaps function

* year - year the composite represents.
* day - the (nominal) day within the year the composite represents (a value of zero and day will not be used)
* compImg - The input compsite image which has been generated.
* imgRef -  The reference layer (e.g., from createMaxNDVIComposite or createMaxNDVINDWICompositeLandsat) with zero for no data regions
* outRef - A boolean variable specify which layer a fill reference layer is to be produced.

"""
    def __init__(self, year=1900, day=0, compImg=None, imgRef=None, outRef=False):
        """
        * year - year the composite represents.
        * day - the (nominal) day within the year the composite represents (a value of zero and day will not be used)
        * compImg - The input compsite image which has been generated.
        * imgRef -  The reference layer (e.g., from createMaxNDVIComposite or createMaxNDVINDWICompositeLandsat) with zero for no data regions
        * outRef - A boolean variable specify which layer a fill reference layer is to be produced.
        """
        self.year = year
        self.day = day
        self.compImg = compImg
        self.imgRef = imgRef
        self.outRef = outRef
    
    def __repr__(self):
        return repr((self.year, self.day, self.compImg, self.imgRef, self.outRef))


def setBandNames(inputImage, bandNames, feedback=False):
    """A utility function to set band names.
Where:

* inImage is the input image
* bandNames is a list of band names
* feedback is a boolean specifying whether feedback will be printed to the console (True= Printed / False (default) Not Printed)

Example::

    from rsgislib import imageutils

    inputImage = 'injune_p142_casi_sub_utm.kea'
    bandNames = ['446nm','530nm','549nm','569nm','598nm','633nm','680nm','696nm','714nm','732nm','741nm','752nm','800nm','838nm']
    
    imageutils.setBandNames(inputImage, bandNames)
    
"""
    dataset = gdal.Open(inputImage, gdal.GA_Update)
    
    for i in range(len(bandNames)):
        band = i+1
        bandName = bandNames[i]

        imgBand = dataset.GetRasterBand(band)
        # Check the image band is available
        if not imgBand is None:
            if feedback:
                print('Setting Band {0} to "{1}"'.format(band, bandName))
            imgBand.SetDescription(bandName)
        else:
            raise Exception("Could not open the image band: ", band)

def getBandNames(inputImage):
    """
A utility function to get band names.

Where:

* inImage is the input image

Return: 

* list of band names

Example::

    from rsgislib import imageutils

    inputImage = 'injune_p142_casi_sub_utm.kea'
    bandNames = imageutils.getBandNames(inputImage)

"""
    dataset = gdal.Open(inputImage, gdal.GA_Update)
    bandNames = list()
    
    for i in range(dataset.RasterCount):
        imgBand = dataset.GetRasterBand(i+1)
        # Check the image band is available
        if not imgBand is None:
            bandNames.append(imgBand.GetDescription())
        else:
            raise Exception("Could not open the image band: ", band)
    return bandNames

def getRSGISLibDataType(inImg):
    """
Returns the rsgislib datatype ENUM for a raster file

* inImg -- The file to get the datatype for

return:
    The rsgislib datatype enum, e.g., rsgislib.TYPE_8INT

"""
    raster = gdal.Open(inImg, gdal.GA_ReadOnly)
    if raster == None:
        raise Exception("Could not open the inImg.")
    band = raster.GetRasterBand(1)
    gdal_dtype = gdal.GetDataTypeName(band.DataType)
    raster = None
    rsgis_utils = rsgislib.RSGISPyUtils()

    return rsgis_utils.getRSGISLibDataType(gdal_dtype)
    
def getGDALDataType(inImg):
    """
Returns the rsgislib datatype ENUM for a raster file

*  inImg -- The file to get the datatype for

return:
    The rsgislib datatype enum, e.g., rsgislib.TYPE_8INT

"""
    raster = gdal.Open(inImg, gdal.GA_ReadOnly)
    if raster == None:
        raise Exception("Could not open the inImg.")
    band = raster.GetRasterBand(1)
    gdal_dtype = gdal.GetDataTypeName(band.DataType)
    raster = None
    return gdal_dtype

def setImgThematic(imageFile):
    """
Set all image bands to be thematic. 

* imageFile - The file for which the bands are to be set as thematic

"""
    ds = gdal.Open(imageFile, gdal.GA_Update)
    if ds == None:
        raise Exception("Could not open the imageFile.")
    for bandnum in range(ds.RasterCount):
        band = ds.GetRasterBand(bandnum + 1)
        band.SetMetadataItem('LAYER_TYPE', 'thematic')
    ds = None


def hasGCPs(inImg):
    """
Test whether the input image has GCPs - returns boolean

* inImg - input image file

Return: 

* boolean True - has GCPs; False - does not have GCPs
"""
    raster = gdal.Open(inImg, gdal.GA_ReadOnly)
    if raster == None:
        raise Exception("Could not open the inImg.")
    numGCPs = raster.GetGCPCount()
    hasGCPs = False
    if numGCPs > 0:
        hasGCPs = True
    raster = None
    return hasGCPs

def copyGCPs(srcImg, destImg):
    """
Copy the GCPs from the srcImg to the destImg

* srcImg - Raster layer with GCPs
* destImg - Raster layer to which GCPs will be added
    
"""
    srcDS = gdal.Open(srcImg, gdal.GA_ReadOnly)
    if srcDS == None:
        raise Exception("Could not open the srcImg.")
    destDS = gdal.Open(destImg, gdal.GA_Update)
    if destDS == None:
        raise Exception("Could not open the destImg.")
        srcDS = None

    numGCPs = srcDS.GetGCPCount()
    if numGCPs > 0:
        gcpProj = srcDS.GetGCPProjection()
        gcpList = srcDS.GetGCPs()
        destDS.SetGCPs(gcpList, gcpProj)

    srcDS = None
    destDS = None

def getWKTProjFromImage(inImg):
    """
A function which returns the WKT string representing the projection of the input image.
* inImg - input image from which WKT string will be read.

"""
    rasterDS = gdal.Open(inImg, gdal.GA_ReadOnly)
    if rasterDS == None:
        raise Exception('Could not open raster image: \'' + inImg+ '\'')
    projStr = rasterDS.GetProjection()
    rasterDS = None
    return projStr

def createBlankImgFromRefVector(inVecFile, inVecLyr, outputImg, outImgRes, outImgNBands, gdalformat, datatype):
    """
A function to create a new image file based on a vector layer to define the extent and projection
of the output image. 

* inVecFile - input vector file.
* inVecLyr - name of the vector layer, if None then assume the layer name will be the same as the file
             name of the input vector file.
* outputImg - output image file.
* outImgRes - output image resolution, square pixels so a single value.
* outImgNBands - the number of image bands in the output image 
* gdalformat - output image file format.
* datatype - output image data type.

"""

    rsgisUtils = rsgislib.RSGISPyUtils()

    baseExtent = rsgisUtils.getVecLayerExtent(inVecFile, inVecLyr)
    xMin, xMax, yMin, yMax = rsgisUtils.findExtentOnGrid(baseExtent, outImgRes, fullContain=True)

    tlX = xMin
    tlY = yMax
    
    widthCoord = xMax - xMin
    heightCoord = yMax - yMin
    
    width = int(math.ceil(widthCoord/outImgRes))
    height = int(math.ceil(heightCoord/outImgRes))
    
    wktString = rsgisUtils.getProjWKTFromVec(inVecFile)

    rsgislib.imageutils.createBlankImage(outputImg, outImgNBands, width, height, tlX, tlY, outImgRes, 0.0, '', wktString, gdalformat, datatype)

def resampleImage2Match(inRefImg, inProcessImg, outImg, gdalformat, interpMethod, datatype=None, noDataVal=None, multicore=False):
    """
A utility function to resample an existing image to the projection and/or pixel size of another image.

Where:

* inRefImg is the input reference image to which the processing image is to resampled to.
* inProcessImg is the image which is to be resampled.
* outImg is the output image file.
* gdalformat is the gdal format for the output image.
* interpMethod is the interpolation method used to resample the image [bilinear, lanczos, cubicspline, nearestneighbour, cubic, average, mode]
* datatype is the rsgislib datatype of the output image (if none then it will be the same as the input file).
* multicore - use multiple processing cores (Default = False)

""" 
    rsgisUtils = rsgislib.RSGISPyUtils()
    numBands = rsgisUtils.getImageBandCount(inProcessImg)
    if noDataVal == None:
        noDataVal = rsgisUtils.getImageNoDataValue(inProcessImg)
    
    if datatype == None:
        datatype = rsgisUtils.getRSGISLibDataTypeFromImg(inProcessImg)
        
    interpolationMethod = gdal.GRA_NearestNeighbour
    if interpMethod == 'bilinear':
        interpolationMethod = gdal.GRA_Bilinear 
    elif interpMethod == 'lanczos':
        interpolationMethod = gdal.GRA_Lanczos 
    elif interpMethod == 'cubicspline':
        interpolationMethod = gdal.GRA_CubicSpline 
    elif interpMethod == 'nearestneighbour':
        interpolationMethod = gdal.GRA_NearestNeighbour 
    elif interpMethod == 'cubic':
        interpolationMethod = gdal.GRA_Cubic
    elif interpMethod == 'average':
        interpolationMethod = gdal.GRA_Average
    elif interpMethod == 'mode':
        interpolationMethod = gdal.GRA_Mode
    else:
        raise Exception("Interpolation method was not recognised or known.")
    
    backVal = 0.0
    haveNoData = False
    if noDataVal != None:
        backVal = float(noDataVal)
        haveNoData = True
    
    rsgislib.imageutils.createCopyImage(inRefImg, outImg, numBands, backVal, gdalformat, datatype)

    inFile = gdal.Open(inProcessImg, gdal.GA_ReadOnly)
    outFile = gdal.Open(outImg, gdal.GA_Update)

    wrpOpts = []
    if multicore:
        if haveNoData:
            wrpOpts = gdal.WarpOptions(resampleAlg=interpolationMethod, srcNodata=noDataVal, dstNodata=noDataVal, multithread=True, callback=gdal.TermProgress)    
        else:
            wrpOpts = gdal.WarpOptions(resampleAlg=interpolationMethod, multithread=True, callback=gdal.TermProgress)
    else:
        if haveNoData:
            wrpOpts = gdal.WarpOptions(resampleAlg=interpolationMethod, srcNodata=noDataVal, dstNodata=noDataVal, multithread=False, callback=gdal.TermProgress)    
        else:
            wrpOpts = gdal.WarpOptions(resampleAlg=interpolationMethod, multithread=False, callback=gdal.TermProgress)
    
    gdal.Warp(outFile, inFile, options=wrpOpts)
    
    inFile = None
    outFile = None


def reprojectImage(inputImage, outputImage, outWKT, gdalformat='KEA', interp='cubic', inWKT=None, noData=0.0, outPxlRes='image', snap2Grid=True, multicore=False):
    """
This function provides a tool which uses the gdalwarp function to reproject an input image.

Where:

* inputImage - the input image name and path
* outputImage - the output image name and path
* outWKT - a WKT file representing the output projection
* gdalformat - the output image file format (Default is KEA)
* interp - interpolation algorithm. Options are: near, bilinear, cubic, cubicspline, lanczos, average, mode. (Default is cubic)
* inWKT - if input image is not well defined this is the input image projection as a WKT file (Default is None, i.e., ignored)
* noData - float representing the not data value (Default is 0.0)
* outPxlRes three inputs can be provided 
            1) 'image' where the output resolution will match the input (Default is image)
            2) 'auto' where an output resolution maintaining the image size of the input image will be used
            3) provide a floating point value for the image resolution (note. pixels will be sqaure) 
* snap2Grid is a boolean specifying whether the TL pixel should be snapped to a multiple of the pixel resolution (Default is True).
* nCores - the number of processing cores available for processing (-1 is all cores: Default=-1)
    """    
    rsgisUtils = rsgislib.RSGISPyUtils()
    
    eResampleAlg = gdal.GRA_CubicSpline
    if interp == 'near':
        eResampleAlg = gdal.GRA_NearestNeighbour
    elif interp == 'bilinear':
        eResampleAlg = gdal.GRA_Bilinear
    elif interp == 'cubic':
        eResampleAlg = gdal.GRA_Cubic
    elif interp == 'cubicspline':
        eResampleAlg = gdal.GRA_CubicSpline
    elif interp == 'lanczos':
        eResampleAlg = gdal.GRA_Lanczos
    elif interp == 'average':
        eResampleAlg = gdal.GRA_Average
    elif interp == 'mode':
        eResampleAlg = gdal.GRA_Mode
    else:
        raise Exception('The interpolation algorithm was not recogonised: \'' + interp + '\'')
    
    if not os.path.exists(inputImage):
        raise Exception('The input image file does not exist: \'' + inputImage + '\'')
    
    inImgDS = gdal.Open(inputImage, gdal.GA_ReadOnly)
    if inImgDS is None:
        raise Exception('Could not open the Input Image: \'' + inputImage + '\'')    
    
    inImgProj = osr.SpatialReference()
    if not inWKT is None:
        if not os.path.exists(inWKT):
            raise Exception('The input WKT file does not exist: \'' + inWKT + '\'')
        inWKTStr = rsgisUtils.readTextFileNoNewLines(inWKT)
        inImgProj.ImportFromWkt(inWKTStr)
    else:
        inImgProj.ImportFromWkt(inImgDS.GetProjectionRef())
        
    if not os.path.exists(outWKT):
        raise Exception('The output WKT file does not exist: \'' + outWKT + '\'')
    outImgProj = osr.SpatialReference()
    outWKTStr = rsgisUtils.readTextFileNoNewLines(outWKT)
    outImgProj.ImportFromWkt(outWKTStr)
    
    geoTransform = inImgDS.GetGeoTransform()
    if geoTransform is None:
        raise Exception('Could read the geotransform from the Input Image: \'' + inputImage + '\'')
    
    xPxlRes = geoTransform[1]
    yPxlRes = geoTransform[5]
    
    inRes = xPxlRes
    if math.fabs(yPxlRes) < math.fabs(xPxlRes):
        inRes = math.fabs(yPxlRes)
    
    xSize = inImgDS.RasterXSize
    ySize = inImgDS.RasterYSize
    
    tlXIn = geoTransform[0]
    tlYIn = geoTransform[3]
    
    brXIn = tlXIn + (xSize * math.fabs(xPxlRes))
    brYIn = tlYIn - (ySize * math.fabs(yPxlRes))
    
    trXIn = brXIn
    trYIn = tlYIn
    
    blXIn = tlXIn
    blYIn = trYIn
    
    numBands = inImgDS.RasterCount
    
    inImgBand = inImgDS.GetRasterBand( 1 );
    gdalDataType = gdal.GetDataTypeName(inImgBand.DataType)
    rsgisDataType = rsgisUtils.getRSGISLibDataType(gdalDataType)
    
    tlPtWKT = 'POINT(%s %s)' % (tlXIn, tlYIn)
    tlPt = ogr.CreateGeometryFromWkt(tlPtWKT)
    tlPt.AssignSpatialReference(inImgProj)
    tlPt.TransformTo(outImgProj)
    tlXOut = tlPt.GetX()
    tlYOut = tlPt.GetY()
    
    brPtWKT = 'POINT(%s %s)' % (brXIn, brYIn)
    brPt = ogr.CreateGeometryFromWkt(brPtWKT)
    brPt.AssignSpatialReference(inImgProj)
    brPt.TransformTo(outImgProj)
    brXOut = brPt.GetX()
    brYOut = brPt.GetY()
    
    trPtWKT = 'POINT(%s %s)' % (trXIn, trYIn)
    trPt = ogr.CreateGeometryFromWkt(trPtWKT)
    trPt.AssignSpatialReference(inImgProj)
    trPt.TransformTo(outImgProj)
    trXOut = trPt.GetX()
    trYOut = trPt.GetY()
    
    blPtWKT = 'POINT(%s %s)' % (blXIn, blYIn)
    btPt = ogr.CreateGeometryFromWkt(blPtWKT)
    btPt.AssignSpatialReference(inImgProj)
    btPt.TransformTo(outImgProj)
    blXOut = btPt.GetX()
    blYOut = btPt.GetY()
    
    xValsOut = [tlXOut, brXOut, trXOut, blXOut]
    yValsOut = [tlYOut, brYOut, trYOut, blYOut]
    
    xMax = max(xValsOut)
    xMin = min(xValsOut)
    
    yMax = max(yValsOut)
    yMin = min(yValsOut)
    
    outPxlRes = str(outPxlRes).strip()
    outRes = 0.0
    if rsgisUtils.isNumber(outPxlRes):
        outRes = math.fabs(float(outPxlRes))
    elif outPxlRes == 'image':
        outRes = inRes
    elif outPxlRes == 'auto':
        xOutRes = (brXOut - tlXOut) / xSize
        yOutRes = (tlYOut - brYOut) / ySize
        outRes = xOutRes
        if yOutRes < xOutRes:
            outRes = yOutRes
    else: 
        raise Exception('Was not able to defined the output resolution. Check Input: \'' + outPxlRes + '\'')
    
    outTLX = xMin
    outTLY = yMax
    outWidth = int(round((xMax - xMin) / outRes)) + 1
    outHeight = int(round((yMax - yMin) / outRes)) + 1
    
    if snap2Grid:
    
        xLeft = outTLX % outRes
        yLeft = outTLY % outRes
        
        outTLX = (outTLX-xLeft) - (5 * outRes)
        outTLY = ((outTLY-yLeft) + outRes) + (5 * outRes)
        
        outWidth = int(round((xMax - xMin) / outRes)) + 10
        outHeight = int(round((yMax - yMin) / outRes)) + 10
    
    print('Creating blank image')
    rsgislib.imageutils.createBlankImage(outputImage, numBands, outWidth, outHeight, outTLX, outTLY, outRes, noData, "", outWKTStr, gdalformat, rsgisDataType)

    outImgDS = gdal.Open(outputImage, gdal.GA_Update)
    
    for i in range(numBands):
        outImgDS.GetRasterBand(i+1).SetNoDataValue(noData)
    
    print("Performing the reprojection")
    
    wrpOpts = []
    if multicore:
        wrpOpts = gdal.WarpOptions(resampleAlg=eResampleAlg, srcNodata=noData, dstNodata=noData, multithread=True, callback=gdal.TermProgress)
    else:
        wrpOpts = gdal.WarpOptions(resampleAlg=eResampleAlg, srcNodata=noData, dstNodata=noData, multithread=False, callback=gdal.TermProgress)    

    gdal.Warp(outImgDS, inImgDS, options=wrpOpts)

    inImgDS = None
    outImgDS = None    



def subsetImgs2CommonExtent(inImagesDict, outShpEnv, gdalformat):
    """
A command to subset a set of images to the same overlapped extent.

Where:

* inImagesDict is a list of dictionaries containing values for IN (input image) OUT (output image) and TYPE (data type for output)
* outShpEnv is a file path for the output shapefile representing the overlap extent.
* gdalformat is the gdal format of the output images.

Example::
    
    from rsgislib import imageutils
    
    inImagesDict = []
    inImagesDict.append({'IN': './Images/Lifeformclip.tif', 'OUT':'./Subsets/Lifeformclip_sub.kea', 'TYPE':rsgislib.TYPE_32INT})
    inImagesDict.append({'IN': './Images/chmclip.tif', 'OUT':'./Subsets/chmclip_sub.kea', 'TYPE':rsgislib.TYPE_32FLOAT})
    inImagesDict.append({'IN': './Images/peakBGclip.tif', 'OUT':'./Subsets/peakBGclip_sub.kea', 'TYPE':rsgislib.TYPE_32FLOAT})
    
    outputVector = 'imgSubExtent.shp'
    imageutils.subsetImgs2CommonExtent(inImagesDict, outputVector, 'KEA')
    
"""
    import rsgislib.vectorutils
    
    inImages = []
    for inImgDict in inImagesDict:
        inImages.append(inImgDict['IN'])
    
    rsgislib.vectorutils.findCommonImgExtent(inImages, outShpEnv, True)
    
    for inImgDict in inImagesDict:
        rsgislib.imageutils.subset(inImgDict['IN'], outShpEnv, inImgDict['OUT'], gdalformat, inImgDict['TYPE'])
    



def buildImgSubDict(globFindImgsStr, outDir, suffix, ext):
    """
Automate building the dictionary of image to be used within the 
subsetImgs2CommonExtent(inImagesDict, outShpEnv, imgFormat) function.

Where:

* globFindImgsStr is a string to be passed to the glob module to find the input image files.
* outDir is the output directory path for the images.
* suffix is a suffix to be appended on to the end of the file name (can be a blank string, i.e., '')
* ext is a string with the output file extension

Example::
    
    from rsgislib import imageutils
    
    inImagesDict = imageutils.buildImgSubDict("./Images/*.tif", "./Subsets/", "_sub", ".kea")
    print(inImagesDict)
    
    outputVector = 'imgSubExtent.shp'
    imageutils.subsetImgs2CommonExtent(inImagesDict, outputVector, 'KEA')

"""
    import glob
    import os.path
        
    inImagesDict = []
    
    inputImages = glob.glob(globFindImgsStr)
    if len(inputImages) == 0:
        raise Exception("No images were found using \'" + globFindImgsStr + "\'")
    
    for image in inputImages:
        dataset = gdal.Open(image, gdal.GA_ReadOnly)
        gdalDType = dataset.GetRasterBand(1).DataType
        dataset = None
        datatype = rsgislib.TYPE_32FLOAT
        if gdalDType == gdal.GDT_Byte:
            datatype = rsgislib.TYPE_8UINT
        elif gdalDType == gdal.GDT_Int16:
            datatype = rsgislib.TYPE_16INT
        elif gdalDType == gdal.GDT_Int32:
            datatype = rsgislib.TYPE_32INT
        elif gdalDType == gdal.GDT_UInt16:
            datatype = rsgislib.TYPE_16UINT
        elif gdalDType == gdal.GDT_UInt32:
            datatype = rsgislib.TYPE_32UINT         
        elif gdalDType == gdal.GDT_Float32:
            datatype = rsgislib.TYPE_32FLOAT
        elif gdalDType == gdal.GDT_Float64:
            datatype = rsgislib.TYPE_64FLOAT
        else:
            raise Exception("Data type of the input file was not recognised or known.")
            
        imgBase = os.path.splitext(os.path.basename(image))[0]
        outImg = os.path.join(outDir, (imgBase+suffix+ext))
        inImagesDict.append({'IN':image, 'OUT':outImg, 'TYPE':datatype})

    return inImagesDict


def calcPixelLocations(inputImg, outputImg, gdalformat):
    """
Function which produces a 2 band output image with the X and Y locations of the image pixels.

Where:

* inputImg - the input reference image
* outputImg - the output image file name and path (will be same dimensions as the input)
* gdalformat - the GDAL image file format of the output image file.

"""
    if not haveRIOS:
        raise Exception("The RIOS module required for this function could not be imported\n\t" + riosErr)

    infiles = applier.FilenameAssociations()
    infiles.image1 = inputImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    aControls = applier.ApplierControls()
    aControls.progress = cuiprogress.CUIProgressBar()
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    
    def _getXYPxlLocs(info, inputs, outputs, otherargs):
        """
        This is an internal rios function 
        """
        xBlock, yBlock = info.getBlockCoordArrays()
        outputs.outimage = numpy.stack((xBlock,yBlock))

    applier.apply(_getXYPxlLocs, infiles, outfiles, otherargs, controls=aControls)

def mergeExtractedHDF5Data(h5Files, outH5File):
    """
A function to merge a list of HDF files (e.g., from rsgislib.imageutils.extractZoneImageBandValues2HDF)
with the same number of variables (i.e., columns) into a single file. For example, if class training
regions have been sourced from multiple images. 

* h5Files - a list of input files. 
* outH5File - the output file.

Example::

    inTrainSamples = ['MSS_CloudTrain1.h5', 'MSS_CloudTrain2.h5', 'MSS_CloudTrain3.h5']
    cloudTrainSamples = 'LandsatMSS_CloudTrainingSamples.h5'
    rsgislib.imageutils.mergeExtractedHDF5Data(inTrainSamples, cloudTrainSamples)

"""
    import h5py
    
    first = True
    numVars = 0
    numVals = 0
    for h5File in h5Files:
        fH5 = h5py.File(h5File)
        dataShp = fH5['DATA/DATA'].shape
        if first:
            numVars = dataShp[1]
            first = False
        elif numVars is not dataShp[1]:
            raise rsgislib.RSGISPyException("The number of variables within the inputted HDF5 files was not the same.")
        numVals += dataShp[0]
        fH5.close()
    
    dataArr = numpy.zeros([numVals, numVars], dtype=float)
    
    rowInit = 0
    for h5File in h5Files:
        fH5 = h5py.File(h5File)
        numRows = fH5['DATA/DATA'].shape[0]
        dataArr[rowInit:(rowInit+numRows)] = fH5['DATA/DATA']
        rowInit += numRows
        fH5.close()
    
    fH5Out = h5py.File(outH5File,'w')
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset('DATA', data=dataArr, chunks=True, compression="gzip", shuffle=True)
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = 'Merged'.encode()
    fH5Out.close()


def doImagesOverlap(image1, image2, overThres=0.0):
    """
Function to test whether two images overlap with one another.
If the images have a difference projection/coordinate system then corners 

* image1 - path to first image
* image2 - path to second image
* overThres - the amount of overlap required to return true (e.g., at least 1 pixel)

Returns:

Boolean specifying whether they overlap or not.

Example::

    import rsgislib.imageutils
    img = "/Users/pete/Temp/LandsatStatsImgs/MSS/ClearSkyMsks/LS1MSS_19720823_lat52lon114_r24p218_osgb_clearsky.tif"
    tile = "/Users/pete/Temp/LandsatStatsImgs/MSS/RefImages/LandsatWalesRegion_60m_tile8.kea"
    
    overlap = rsgislib.imageutils.doImagesOverlap(tile, img)
    print("Images Overlap: " + str(overlap))

"""
    overlap = True
    
    projSame = False
    rsgisUtils = rsgislib.RSGISPyUtils()
    if rsgisUtils.doGDALLayersHaveSameProj(image1, image2):
        projSame = True
    
    img1DS = gdal.Open(image1, gdal.GA_ReadOnly)
    if img1DS is None:
        raise rsgislib.RSGISPyException('Could not open image: ' + image1)
        
    img2DS = gdal.Open(image2, gdal.GA_ReadOnly)
    if img2DS is None:
        raise rsgislib.RSGISPyException('Could not open image: ' + image2)

    img1GeoTransform = img1DS.GetGeoTransform()
    if img1GeoTransform is None:
        img1DS = None
        img2DS = None
        raise rsgislib.RSGISPyException('Could not get geotransform: ' + image1)
        
    img2GeoTransform = img2DS.GetGeoTransform()
    if img2GeoTransform is None:
        img1DS = None
        img2DS = None
        raise rsgislib.RSGISPyException('Could not get geotransform: ' + image2)
    
    img1TLX = img1GeoTransform[0]
    img1TLY = img1GeoTransform[3]
    
    img1BRX = img1GeoTransform[0] + (img1DS.RasterXSize * img1GeoTransform[1])
    img1BRY = img1GeoTransform[3] + (img1DS.RasterYSize * img1GeoTransform[5])
    
    img2TLX_orig = img2GeoTransform[0]
    img2TLY_orig = img2GeoTransform[3]
    
    img2BRX_orig = img2GeoTransform[0] + (img2DS.RasterXSize * img2GeoTransform[1])
    img2BRY_orig = img2GeoTransform[3] + (img2DS.RasterYSize * img2GeoTransform[5])
    
    img1EPSG = rsgisUtils.getEPSGCode(image1)
    img2EPSG = rsgisUtils.getEPSGCode(image2)
    
    if projSame:
        img2TLX = img2GeoTransform[0]
        img2TLY = img2GeoTransform[3]
        
        img2BRX = img2GeoTransform[0] + (img2DS.RasterXSize * img2GeoTransform[1])
        img2BRY = img2GeoTransform[3] + (img2DS.RasterYSize * img2GeoTransform[5])
    else:
        inProj = osr.SpatialReference()
        
        if img2EPSG is None:
            wktImg2 = rsgisUtils.getWKTProjFromImage(image2)
            if (wktImg2 is None) or (wktImg2 == ""):
                raise rsgislib.RSGISPyException('Could not retrieve EPSG or WKT for image: ' + image2)
            inProj.ImportFromWkt(wktImg2)
        else:
            inProj.ImportFromEPSG(int(img2EPSG))
        
        outProj = osr.SpatialReference()
        if img1EPSG is None:
            wktImg1 = rsgisUtils.getWKTProjFromImage(image1)
            if (wktImg1 is None) or (wktImg1 == ""):
                raise rsgislib.RSGISPyException('Could not retrieve EPSG or WKT for image: ' + image1)
            outProj.ImportFromWkt(wktImg1)
        else:
            outProj.ImportFromEPSG(int(img1EPSG))
        
        if img1EPSG is None:
            img1EPSG = 0

        if img1EPSG == 4326:
            img2TLY, img2TLX = rsgisUtils.reprojPoint(inProj, outProj, img2TLX_orig, img2TLY_orig)
            img2BRY, img2BRX = rsgisUtils.reprojPoint(inProj, outProj, img2BRX_orig, img2BRY_orig)
        else:
            img2TLX, img2TLY = rsgisUtils.reprojPoint(inProj, outProj, img2TLX_orig, img2TLY_orig)
            img2BRX, img2BRY = rsgisUtils.reprojPoint(inProj, outProj, img2BRX_orig, img2BRY_orig)
    
    xMin = img1TLX
    xMax = img1BRX
    yMin = img1BRY
    yMax = img1TLY
    
    if img2TLX > xMin:
        xMin = img2TLX
    if img2BRX < xMax:
        xMax = img2BRX
    if img2BRY > yMin:
        yMin = img2BRY
    if img2TLY < yMax:
        yMax = img2TLY
        
    if xMax - xMin <= overThres:
        overlap = False
    elif yMax - yMin <= overThres:
        overlap = False

    return overlap


def generateRandomPxlValsImg(inputImg, outputImg, gdalformat, lowVal, upVal):
    """
Function which produces a 1 band image with random values between lowVal and upVal.

Where:

* inputImg - the input reference image
* outputImg - the output image file name and path (will be same dimensions as the input)
* gdalformat - the GDAL image file format of the output image file.
* lowVal - lower value
* upVal - upper value 

"""
    if not haveRIOS:
        raise Exception("The RIOS module required for this function could not be imported\n\t" + riosErr)

    infiles = applier.FilenameAssociations()
    infiles.inImg = inputImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    otherargs.lowVal = lowVal
    otherargs.upVal = upVal
    aControls = applier.ApplierControls()
    aControls.progress = cuiprogress.CUIProgressBar()
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    
    def _popPxlsRanVals(info, inputs, outputs, otherargs):
        """
        This is an internal rios function for generateRandomPxlValsImg()
        """
        outputs.outimage = numpy.random.random_integers(otherargs.lowVal, high=otherargs.upVal, size=inputs.inImg.shape)
        outputs.outimage = outputs.outimage.astype(numpy.int32, copy=False)
    
    applier.apply(_popPxlsRanVals, infiles, outfiles, otherargs, controls=aControls)


def extractImgPxlSample(inputImg, pxlNSample, noData=None):
    """
A function which extracts a sample of pixels from the 
input image file to a number array.

* inputImg - the image from which the random sample will be taken.
* pxlNSample - the sample to be taken (e.g., a value of 100 will sample every 100th, valid (if noData specified), pixel)
* noData - provide a no data value which is to be ignored during processing. If None then ignored (Default: None)

return:: outputs a numpy array (n sampled values, n bands)

""" 
    # Import the RIOS image reader
    from rios.imagereader import ImageReader
    
    first = True
    reader = ImageReader(inputImg, windowxsize=200, windowysize=200)
    print('Started .0.', end='', flush=True)
    outCount = 10
    for (info, block) in reader:
        if info.getPercent() > outCount:
            print('.'+str(int(outCount))+'.', end='', flush=True)
            outCount = outCount + 10
                
        blkShape = block.shape
        blkBands = block.reshape((blkShape[0], (blkShape[1]*blkShape[2])))
        
        blkBandsTrans = numpy.transpose(blkBands)
        
        if noData is not None:
            blkBandsTrans = blkBandsTrans[(blkBandsTrans!=noData).all(axis=1)]
        
        if blkBandsTrans.shape[0] > 0:
            nSamp = int((blkBandsTrans.shape[0])/pxlNSample)
            nSampRange = numpy.arange(0, nSamp, 1)*pxlNSample
            blkBandsTransSamp = blkBandsTrans[nSampRange]
            
            if first:
                outArr = blkBandsTransSamp
                first = False
            else:
                outArr = numpy.concatenate((outArr, blkBandsTransSamp), axis=0)
    print('. Completed')
    return outArr





#!/usr/bin/env python

#############################################
# rsgislibmosaic.py
#
#  Copyright 2014 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose: A script to recursivly find files with a given search
# string and mosaic using RGISLib.
#
# Author: Dan Clewley
# Email: daniel.clewley@gmail.com
# Date: 28/08/2013
# Version: 1.0
# 
##############################################

import os, sys
import rsgislib
from rsgislib import imageutils
import argparse
import fnmatch

"""
# Print help
print('rsgislibmosaic.py script provides command line utility for')
print('mosaicking image files, including recursively finding files')
print('within an input directory.')
print('This script was distributed with RSGISLib 3.6.14')
print('Copyright (C) 2014 Peter Bunting and Daniel Clewley')
print('For support please email rsgislib-support@googlegroups.com\n')
"""

# Get input parameters
parser = argparse.ArgumentParser()
parser.add_argument("--inputimages", type=str, nargs='+', required=False, help="Provide a list of images to be mosaicked rather than search info.")
parser.add_argument("-i", "--indir", type=str, required=False, help="Input directory to recursively search")
parser.add_argument("-s", "--search", type=str, required=False, help="Search string, e.g., '*kea', must be in quotes.")
parser.add_argument("-o", "--outmosaic", type=str, required=True, help="Output mosaic file")
parser.add_argument("-l",'--outlist',type=str, default=None,help="Output text file with list of files in mosaic (optional)")
parser.add_argument("-ot",'--datatype',choices=['int8', 'int16', 'int32', 'int64', 'uint8', 'byte', 'uint16', 'uint32', 'uint64', 'float32', 'float64'], default='float32',help="Data type")
parser.add_argument("--backgroundval", type=float, default=0,help="Background Value (default 0)")
parser.add_argument("--skipval", type=float, default=0,help="Value to be skipped (nodata values) in the input images (default 0)")
parser.add_argument("--skipband", type=int, default=1,help="Band to check for skip val (default 1)")
parser.add_argument("--minpix", action='store_true', default=False, help="Use minimum pixel in overlap areas (default, use last image in)")
parser.add_argument("--maxpix", action='store_true', default=False, help="Use maximum pixel in overlap areas (default, use last image in)")
parser.add_argument("--nostats", action='store_true', default=False, help="Don't calculate statistics and pyramids for mosaic (default is to calculate)")
args = parser.parse_args()    

rsgisUtils = rsgislib.RSGISPyUtils()

search4InImgs = True
if args.inputimages != None:
    if len(args.inputimages) > 0:
        search4InImgs = False


if search4InImgs and ((args.indir == "") or (args.search == "")):
    print('ERROR: Either a list of input images (--inputimages) or a search path (--indir) and string (--search) did to be given.')
    sys.exit()


overlapBehaviour = 0
if args.minpix and args.maxpix:
    print("ERROR: Either '--minpix' or '--maxpix' expected (not both)")
    sys.exit()
elif args.minpix:
    print("Taking minimum pixel value in band {} for overlapping areas.".format(args.skipband))
    overlapBehaviour = 1
elif args.maxpix:
    print("Taking maximum pixel value in band {} for overlapping areas.".format(args.skipband))
    overlapBehaviour = 2
else:
    print("Using values of last image for overlapping areas.")

# Get output extension from input file
outFormat = rsgisUtils.getGDALFormatFromExt(args.outmosaic)

if search4InImgs:
    # Recursively find all files in the input directory using specified search string
    fileList = []
    
    # Walk through directory
    for dName, sdName, fList in os.walk(args.indir):
        for fileName in fList:
            if fnmatch.fnmatch(fileName, args.search): # Match search string
                fileList.append(os.path.join(dName, fileName))
else:
    fileList = args.inputimages

fileCount=len(fileList)
if fileCount == 0:
    print('ERROR: No files found')
    sys.exit()
else:
    print('Found %i files'%fileCount)

# Save list of files
if args.outlist is not None:
    rsgisUtils.writeList2File(fileList, args.outlist)

print('Creating mosaic...')
t = rsgislib.RSGISTime()
t.start(True)
imageutils.createImageMosaic(fileList, args.outmosaic, args.backgroundval, args.skipval, args.skipband, overlapBehaviour, outFormat, rsgisUtils.getRSGISLibDataType(args.datatype))
t.end()

if not args.nostats:
    print('\nCalculating stats and pyramids...')
    t.start(True)
    imageutils.popImageStats(args.outmosaic, True, args.backgroundval, True)
    t.end()
print('rsgismosaic.py - Finished')
    



"""
The tools.projection module contains some useful tools for working with projections.
"""

# Import numpy
import numpy

haveGDALPy = True
try:
    import osgeo.gdal as gdal
except ImportError as gdalErr:
    haveGDALPy = False
   
def degrees_to_metres(latitude, lonsize, latsize):
    """ 
Convert pixel sizes or distances in degrees to metres.

Where:

* latitude - latitude
* lonsize - numpy array of x pixel sizes (degrees)
* latsize - numpy array of y pixel sizes (degrees)

Returns:

* xsize - numpy array of x pixel sizes (m)
* ysize - numpy array of y pixel sizes (m)

Example::

    from rsgislib import tools
    xsize, ysize = tools.degrees_to_metres(52,1.0,1.0)
    

    """

    # Set up parameters for ellipse
    # Semi-major and semi-minor for WGS-84 ellipse
    ellipse = [6378137.0, 6356752.314245]
    
    radlat = numpy.deg2rad(latitude)
    
    Rsq = (ellipse[0]*numpy.cos(radlat))**2+(ellipse[1]*numpy.sin(radlat))**2
    Mlat = (ellipse[0]*ellipse[1])**2/(Rsq**1.5)
    Nlon = ellipse[0]**2/numpy.sqrt(Rsq)
    xsize = numpy.pi/180*numpy.cos(radlat)*Nlon*lonsize
    ysize = numpy.pi/180*Mlat*latsize

    return xsize, ysize

def metres_to_degrees(latitude, xsize, ysize):
    """ 
Convert pixel sizes or distances in metres to degrees.

Where:

* latitude - latitude
* xsize - numpy array of x pixel sizes (m)
* ysize - numpy array of y pixel sizes (m)

Returns:

* lonsize - numpy array of x pixel sizes (degrees)
* latsize - numpy array of y pixel sizes (degrees)

Example::

    from rsgislib import tools
    xsize, ysize = tools.metres_to_degrees(52,1.0,1.0)
    
    """

    # Set up parameters for ellipse
    # Semi-major and semi-minor for WGS-84 ellipse
    ellipse = [6378137.0, 6356752.314245]
    
    radlat = numpy.deg2rad(latitude)
    
    Rsq = (ellipse[0]*numpy.cos(radlat))**2+(ellipse[1]*numpy.sin(radlat))**2
    Mlat = (ellipse[0]*ellipse[1])**2/(Rsq**1.5)
    Nlon = ellipse[0]**2/numpy.sqrt(Rsq)
    lonsize = xsize / (numpy.pi/180*numpy.cos(radlat)*Nlon)
    latsize = ysize / (numpy.pi/180*Mlat)

    return lonsize, latsize





"""
The zonal stats module provides functions to perform pixel-in-polygon or point-in-pixel analysis. 
"""

# import the C++ extension into this level
from ._zonalstats import *

METHOD_POLYCONTAINSPIXEL = 0           # Polygon completely contains pixel
METHOD_POLYCONTAINSPIXELCENTER = 1     # Pixel center is within the polygon
METHOD_POLYOVERLAPSPIXEL = 2           # Polygon overlaps the pixel
METHOD_POLYOVERLAPSORCONTAINSPIXEL = 3 # Polygon overlaps or contains the pixel
METHOD_PIXELCONTAINSPOLY = 4           # Pixel contains the polygon
METHOD_PIXELCONTAINSPOLYCENTER = 5     # Polygon center is within pixel
METHOD_ADAPTIVE = 6                    # The method is chosen based on relative areas of pixel and polygon.
METHOD_ENVELOPE = 7                    # All pixels in polygon envelope chosen
METHOD_PIXELAREAINPOLY = 8             # Percent of pixel area that is within the polygon
METHOD_POLYAREAINPIXEL = 9             # Percent of polygon area that is within pixel

class ZonalAttributes:
    """ Object, specifying which stats should be calculated and minimum / maximum thresholds. 
This is passed to the pixelStats2SHP and pixelStats2TXT functions. """
    def __init__(self, minThreshold=None, maxThreshold=None, calcCount=False, calcMin=False, calcMax=False, calcMean=False, calcStdDev=False, calcMode=False, calcSum=False):
        self.minThreshold = minThreshold
        self.maxThreshold = maxThreshold
        self.calcCount = calcCount
        self.calcMin = calcMin
        self.calcMax = calcMax
        self.calcMean = calcMean
        self.calcStdDev = calcStdDev 
        self.calcMode = calcMode
        self.calcSum = calcSum
